import { ChangeSet } from "@codemirror/state";
import { FakeSelection } from "./fake-selection.mjs";
export { fakeSelection } from "./fake-selection.mjs";
/** Reserved command for specifying file. */
export const selectCmd = "file:";
/** Reserved string for specifying scroll actions. */
export const scrollCmd = "s";
/**
 * Replay typing in CodeMirror.
 * @returns Unsubscription function.
 */
export function cmReplay({ data, handle, playback, scrollBehavior, shouldScroll = () => true, start, view, }) {
    return cmReplayMultiple({
        data: [[0, `${selectCmd}default`], ...data],
        handle: (key, docs) => handle(key, docs.default),
        playback,
        scrollBehavior,
        shouldScroll,
        start,
        views: {
            default: view,
        },
    });
}
/**
 * Replay typing to several CodeMirror instances in parallel.
 * @returns Unsubscription function.
 */
export function cmReplayMultiple({ data, handle, playback, scrollBehavior = "auto", shouldScroll = () => true, start = 0, views, }) {
    /** Current file being replayed into */
    let file = undefined;
    // validation
    if (!(data.length > 0 &&
        data[0][0] === 0 &&
        typeof data[0][1] === "string" &&
        data[0][1].startsWith(selectCmd))) {
        throw new Error("First command must have time 0 and select the file");
    }
    // we're going to mess with data, clone it
    data = JSON.parse(JSON.stringify(data));
    /* unpackage */
    // decompress times
    const times = data.map((_) => _[0]);
    for (let i = 1; i < times.length; ++i)
        times[i] += times[i - 1];
    // deserialize changesets
    for (const [, action] of data) {
        // changeset
        if (Array.isArray(action) && Array.isArray(action[0])) {
            action[0] = ChangeSet.fromJSON(action[0]);
        }
    }
    // for scrolling in legacy recordings
    const hasScroll = {};
    for (const key in views) {
        hasScroll[key] = false;
    }
    // initialize inverses
    const inverses = {};
    const lastScroll = {};
    for (const key in views) {
        inverses[key] = [];
        lastScroll[key] = [0, 0];
    }
    // compute inverses
    {
        const docs = {};
        for (const key in views) {
            docs[key] = views[key].state.doc;
        }
        for (let i = 0; i < data.length; ++i) {
            const action = data[i][1];
            if (Array.isArray(action)) {
                if (action[0] instanceof ChangeSet) {
                    // editor change
                    inverses[file][i] = action[0].invert(docs[file]);
                    docs[file] = action[0].apply(docs[file]);
                }
                else if (action[0] === scrollCmd) {
                    // scroll
                    hasScroll[file] = true;
                    inverses[file][i] = lastScroll[file];
                    lastScroll[file] = [action[1], action[2] ?? 0];
                }
            }
            else if (action.startsWith(selectCmd)) {
                file = action.slice(selectCmd.length);
            }
        }
    }
    /* main logic */
    let index = 0;
    let lastTime = 0;
    const repaint = () => {
        const t = playback.currentTime;
        const progress = (t - start) * 1000;
        const changes = {};
        for (const key in views) {
            changes[key] = ChangeSet.empty(views[key].state.doc.length);
        }
        const selections = {};
        // apply / revert changes
        if (lastTime <= t && index < data.length) {
            // forward
            let i = index;
            for (; i < data.length && times[i] <= progress; ++i) {
                const action = data[i][1];
                if (typeof action === "string") {
                    if (action.startsWith(selectCmd)) {
                        file = action.slice(selectCmd.length);
                    }
                    // handle action
                    const docs = {};
                    for (const key in views) {
                        docs[key] = changes[key].apply(views[key].state.doc);
                    }
                    handle(action, docs);
                }
                else {
                    if (action[0] === scrollCmd) {
                        if (shouldScroll(file)) {
                            // scroll
                            const [, y, x = 0] = action;
                            const fontSize = getFontSize(views[file]);
                            views[file].scrollDOM.scrollTo({
                                left: x * fontSize,
                                top: y * fontSize,
                                behavior: scrollBehavior,
                            });
                        }
                    }
                    else {
                        // editor change
                        changes[file] = changes[file].compose(action[0]);
                        // handle selection
                        if (action[1]) {
                            const [anchor, head] = action[1];
                            selections[file] = { anchor, head };
                        }
                    }
                }
            }
            index = i;
        }
        else if (t < lastTime && 0 < index) {
            // revert
            let i = index - 1;
            for (; 0 <= i && progress < times[i]; --i) {
                if (inverses[file][i]) {
                    const inverse = inverses[file][i];
                    // editor change
                    if (inverse instanceof ChangeSet) {
                        changes[file] = changes[file].compose(inverses[file][i]);
                    }
                    // scroll
                    else if (inverses[file][i].length === 2) {
                        if (shouldScroll(file)) {
                            const [y, x] = inverses[file][i];
                            const fontSize = getFontSize(views[file]);
                            views[file].scrollDOM.scrollTo({
                                left: x * fontSize,
                                top: y * fontSize,
                                behavior: scrollBehavior,
                            });
                        }
                    }
                }
                else if (data[i][1] === selectCmd + file) {
                    // find file to replay into
                    for (let j = i - 1; 0 <= j; --j) {
                        const action = data[j][1];
                        if (typeof action === "string" && action.startsWith(selectCmd)) {
                            file = action.slice(selectCmd.length);
                            // XXX figure out how to handle more general actions
                            handle(action, {});
                            break;
                        }
                    }
                }
            }
            index = i + 1;
        }
        for (const key in views) {
            const effects = selections[key]
                ? [FakeSelection.of(selections[key])]
                : undefined;
            const view = views[key];
            view.dispatch(view.state.update({
                changes: changes[key],
                effects,
            }));
            // scrolling for legacy recordings
            const scrollIntoView = !hasScroll[key] && shouldScroll(key);
            if (scrollIntoView) {
                // get position of last change
                let pos;
                changes[key].iterChangedRanges((_fromA, _toA, _fromB, toB) => {
                    pos = toB;
                });
                // changes can be empty
                if (pos === undefined) {
                    return;
                }
                const { scrollDOM } = view;
                const rect = scrollDOM.getBoundingClientRect();
                // it isn't possible to measure things that are offscreen
                const line = view.state.doc.lineAt(pos);
                const wordTop = view.defaultLineHeight * (line.number - 1);
                if (wordTop < scrollDOM.scrollTop) {
                    scrollDOM.scrollTo({
                        behavior: scrollBehavior,
                        top: wordTop,
                    });
                }
                else if (wordTop > scrollDOM.scrollTop + rect.height) {
                    scrollDOM.scrollTo({
                        behavior: scrollBehavior,
                        top: wordTop - rect.height + view.defaultLineHeight,
                    });
                }
            }
        }
        lastTime = t;
    };
    /* subscribe */
    playback.addEventListener("seeking", repaint);
    playback.addEventListener("timeupdate", repaint);
    return () => {
        playback.removeEventListener("seeking", repaint);
        playback.removeEventListener("timeupdate", repaint);
    };
}
/** Get the fontSize of a view's scrollDOM. */
function getFontSize(view) {
    return Number.parseFloat(getComputedStyle(view.scrollDOM).fontSize);
}
