import { jsx as _jsx, Fragment as _Fragment } from "react/jsx-runtime";
import { EditorView, ViewPlugin, keymap } from "@codemirror/view";
import { ReplayDataRecorder } from "@liqvid/recording";
import { bind } from "@liqvid/utils/misc";
import { icon } from "./icon.mjs";
import { scrollCmd } from "./index.mjs";
// the actual thingy that gets exported
export class CodeRecorder extends ReplayDataRecorder {
    constructor() {
        super();
        bind(this, ["extension"]);
    }
    /**
     * Get a CodeMirror extension for recording.
     * @param specialKeys Map of key sequences to commands, e.g. {"Mod-Enter": "run"}.
     * @returns CodeMirror extension.
     */
    extension(specialKeys = {}) {
        // legacy
        if (Array.isArray(specialKeys)) {
            // biome-ignore lint/style/noParameterAssign: helpful for backwards compatibility
            specialKeys = Object.fromEntries(specialKeys.map((key) => [key, key]));
        }
        // eslint-disable-next-line @typescript-eslint/no-this-alias
        const $this = this;
        const scrollListener = ViewPlugin.fromClass(class {
            constructor(view) {
                view.scrollDOM.addEventListener("scroll", () => {
                    if (!$this.manager || $this.manager.paused || !$this.manager.active)
                        return;
                    const time = $this.manager.getTime();
                    const fontSize = Number.parseFloat(getComputedStyle(view.scrollDOM).getPropertyValue("font-size"));
                    // vertical scroll is more common so we put it first and omit
                    // horizontal scroll if it's zero
                    const action = [
                        scrollCmd,
                        view.scrollDOM.scrollTop / fontSize,
                        view.scrollDOM.scrollLeft / fontSize,
                    ];
                    if (action[2] === 0) {
                        action.pop();
                    }
                    $this.capture(time, action);
                });
            }
        });
        // record document changes
        const updateListener = EditorView.updateListener.of((update) => {
            if (!this.manager || this.manager.paused || !this.manager.active)
                return;
            // get selection change (if any)
            const transactions = update.transactions
                .map((t) => {
                if (!t.selection)
                    return null;
                const range = t.selection.ranges[0];
                return [range.anchor, range.head];
            })
                .filter(Boolean)
                .slice(-1);
            // empty events can break replay
            if (update.changes.empty && transactions.length === 0)
                return;
            this.capture(this.manager.getTime(), [
                update.changes.toJSON(),
                ...transactions,
                // biome-ignore lint/suspicious/noExplicitAny: TODO
            ]);
        });
        // record special key presses
        const keyListener = keymap.of(Object.keys(specialKeys).map((key) => ({
            key,
            run: () => {
                if (this.manager?.active && !this.manager.paused) {
                    this.capture(this.manager.getTime(), specialKeys[key]);
                }
                return false;
            },
        })));
        return [scrollListener, updateListener, keyListener];
    }
}
const KeySaveComponent = (props) => {
    return (_jsx(_Fragment, { children: _jsx("textarea", { readOnly: true, value: JSON.stringify(props.data) }) }));
};
export const CodeRecording = {
    enabled: () => true,
    icon,
    key: "codemirror",
    name: "Code",
    recorder: new CodeRecorder(),
    saveComponent: KeySaveComponent,
    title: "Record code",
};
